package texastea;

import edu.princeton.cs.*;
import java.awt.Color;

/**
 * <p>
 * I have adapted this code from code
 * included in "Introduction to Programming
 * in Java: An Interdisciplinary Approach," by
 * Robert Sedgewick and Kevin Wayne.
 * </p>
 * 
 * <p>
 * I have repackaged the code from our textbook
 * to make a NetBeans project.
 * </p>
 * 
 * <p>
 * I have also chosen to make less use of static methods.
 * </p>
 * 
 * @author Leon Tabak
 * @version 4 February 2013
 */
public class Percolation {
  private boolean [][] open;
  private boolean [][] full;
  private boolean percolates;
  
  public Percolation( String fileName) {
    In in = new In( "src/texastea/" + fileName );
    int rows = in.readInt();
    int cols = in.readInt();
    this.open = new boolean[rows][cols];
    for (int i = 0; i < rows; i++) {
      for (int j = 0; j < cols; j++) {
        int zeroOrOne = in.readInt();
        if( zeroOrOne == 1 ) {
        this.open[i][j] = true;
        } // if
        else {
          this.open[i][j] = false;
        } // else
      } // for
    } // for
    
    this.full = this.flow();
    this.percolates = this.percolates();
  } // Percolation()
  
  public Percolation( int N, double p  ) {
    this.open = this.random( N, p );
    this.full = this.flow();
    this.percolates = this.percolates();
  } // Percolation( int )
  
  // stub method (Program 2.4.1, page 290)
  private boolean[][] flowStub() {
    int N = this.open.length;
    boolean [][] fullCells = new boolean[N][N];
    // Percolation flow computation goes here.
    return fullCells;
  } // flowStub()

  // vertical percolation detection (Program 2.4.2, page 292)
  private boolean [][] flowVertical() {
    // Compute full sites for vertical percolation
    int N = this.open.length;
    
    boolean [][] fullCells = new boolean[N][N];
    
    for( int j = 0; j < N; j++ ) {
      fullCells[0][j] = this.open[0][j];
    } // for
    
    for( int i = 1; i < N; i++ ) {
      for( int j = 0; j < N; j++ ) {
        fullCells[i][j] = this.open[i][j] && fullCells[i-1][j];
      } // for
    } // for
    
    return fullCells;
  } // flowVertical()
  
  private boolean [][] flow() {
    // Fill every site reachable from the top row.
    int N = this.open.length;
    boolean [][] fullCells = new boolean[N][N];
    for( int j = 0; j < N; j++ ) {
      flow( fullCells, 0, j );
    } // for
    return fullCells;
  } // flow()
  
  private void flow( boolean [][] fullCells, 
          int i, int j ) {
    // Fill every site reachable from (i, j )
    int N = fullCells.length;
    if( i < 0 || i >= N ) { return; }
    if( j < 0 || j >= N ) { return; }
    if( !open[i][j] ) { return; }
    if( fullCells[i][j] ) { return; }
    fullCells[i][j] = true;
    flow( fullCells, i + 1, j ); // Down.
    flow( fullCells, i, j + 1 ); // Right.
    flow( fullCells, i, j - 1 ); // Left.
    flow( fullCells, i - 1, j ); // Up.
  } // flow( boolean [][], boolean [][], int, int )
  
  
  private boolean percolates() {
    int N = this.full.length;
    for (int j = 0; j < N; j++) {
      if (this.full[N - 1][j]) {
        return true;
      } // if
    } // for
    return false;
  } // percolates()

  public boolean [][] getOpen() {
    return this.open;
  } // getOpen()
  
  public boolean [][] getFull() {
    return this.full;
  } // getFull()
  
  public boolean getPercolates() {
    return this.percolates;
  } // getPercolates()
  
  public void showOpenCells( boolean which ) {
    this.show( this.getOpen(), which );
  } // showOpenCells( boolean )
  
  public void showFullCells( boolean which ) {
    this.show( this.getFull(), which );
  } // showFullCells( boolean )
  
  private void show( boolean [][] a, boolean which ) {
    int N = a.length;
    StdDraw.setXscale( -1.0, N );
    StdDraw.setYscale( -1.0, N );
    for( int i = 0; i < N; i++ ) {
      for( int j = 0; j < N; j++ ) {
        if( a[i][j] == which ) {
          StdDraw.filledSquare( j, N - i - 1, 0.5 );
        } // if
      } // for
    } // for
  } // show( boolean [][], boolean )
  
  private boolean [][] random( int N, double p ) {
    boolean [][] a = new boolean[N][N];
    for( int i = 0; i < N; i++ ) {
      for( int j = 0; j < N; j++ ) {
        a[i][j] = StdRandom.bernoulli(p);
      } // for
    } // for
    return a;
  } // random( int, double )
  
  public static void visualize( int N, double p, int T ) {
    for( int t = 0; t < T; t++ ) {
      Percolation percolation = new Percolation( N, p );
      StdDraw.clear();
      StdDraw.setPenColor( Color.BLACK );
      percolation.showOpenCells( true );
      StdDraw.setPenColor( Color.BLUE );
      percolation.showFullCells( true );
      StdDraw.show(1000);
    } // for
  } // visualize()
  
  public static double estimate( int N, double p, int T ) {
    // Generate T random networks, return empirical
    // percolation probability estimate.
    int count = 0;
    for( int t  = 0; t < T; t++ ) {
      // Generate one random network.
      Percolation percolation = new Percolation( N, p );
      if( percolation.percolates() ) {
        count++;
      } // if
    } // for
    return ((double) count)/T ;
  } // estimate()
  
  public static void curve( int N ) {
    Percolation.curve( N, 0.0, 0.0, 1.0, 1.0 );
  } // curve( int )
  
  private static void curve( int N, double x0, double y0, 
          double x1, double y1 ) {
    // Perform experiments and plot results.
    double gap = 0.005;
    double err = 0.5;
    int T = 10000;
    double xm = (x0 + x1)/2;
    double ym = (y0 + y1)/2;
    double fxm = Percolation.estimate( N, xm, T );
    if( (x1 - x0) < gap && Math.abs(ym - fxm) < err ) {
      StdDraw.line( x0, y0, x1, y1);
      return;
    } // if
    Percolation.curve( N, x0, y0, xm, fxm );
    StdDraw.filledCircle( xm, fxm, 0.005 );
    Percolation.curve( N, xm, fxm, x1, y1 );
  } // curve( int, double, double, double, double )
  
  public static void main(String[] args) {
    Percolation p = new Percolation( "test8.txt" );

    // See the medium represented by an array
    // of zeroes and ones.
    // p.flow() shows the cells that water
    // reaches.
    StdArrayIO.print(p.flow());
    StdOut.println(p.percolates());
    
    // See the medium represented by an array
    // of black and white squares.
    // p.getOpen() shows the voids in the
    // medium---water will fill some of the
    // voids but there may also be voids that
    // are not connected to other voids, and
    // so do not fill with water.
    p.show( p.getOpen(), true);

    // Create 8 x 8 arrays in which each cell has
    // a 0.7 probability of being open---do this 24 times.
    // Display images of each of the 24 arrays.
//    Percolation.visualize( 8, 0.7, 24 );
    
    // Create 8 x 8 arrays in which each cell has
    // a 0.7 probability of being open---do this 24 times.
    // Count how many of the 24 arrays contain a network
    // of open cells that allows percolation through the
    // simulated medium---that count, divided by 24, is
    // an estimate of the probability of percolation.
//    double probability = Percolation.estimate( 8, 0.7, 24 );
//    System.out.println( "probability = " + probability );

    // Plot the probability of percolation as a function
    // of the probability that a cell is open for 8 x 8 
    // arrays.
//    Percolation.curve( 8 );
  } // main( String [] )
} // Percolation
